/*
 * block.c -- find/create block devices and their fs nodes
 *
 * Copyright 2006-2008 Red Hat, Inc.  All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Author(s): Peter Jones <pjones@redhat.com>
 */

#define _GNU_SOURCE 1

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <stdlib.h>
#include <stddef.h>

#include <argz.h>
#include <envz.h>

#include <blkid/blkid.h>

#include <linux/blkpg.h>

#include <nash.h>
#include "lib.h"
#include "block.h"
#include "dm.h"

int
nashBdevRemovable(struct nash_block_dev *bdev)
{
    char *subpath = NULL;
    char *contents = NULL;
    int rc;
    int fd;

    rc = asprintfa(&subpath, "%s/removable", bdev->sysfs_path);
    if (rc == -1)
        return 0;

    fd = open(subpath, O_RDONLY);
    if (fd == -1)
        return 0;

    rc = readFD(fd, &contents);
    if (rc < 1) {
        close(fd);
        return 0;
    }

    rc = contents[0] > '0';
    free(contents);
    close(fd);
    return rc;
}

int
nashParseSysfsDevno(const char *path, dev_t *dev)
{
    char *first = NULL, *second;
    int major, minor;
    int fd, len = strlen(path);
    char devname[strlen(path) + 5];
    char devbuf[16];

    nassert(path && path[0]);
    strcpy(devname, path);
    strcat(devname, "/dev");
    fd = open(devname, O_RDONLY);
    if (fd < 0) {
        if (errno == ENOTDIR)
            errno = ENOENT;
        return fd;
    }

    /* leave a space for the NUL character at the end. */
    len = sizeof (devbuf) - 2;
    len = readFDbuf(fd, devbuf, len);
    devbuf[len] = '\0';
    first = devbuf;

    close(fd);

    second = strchr(first, ':');
    *second++ = '\0';

    errno = 0;
    major = strtol(first, NULL, 10);

    errno = 0;
    minor = strtol(second, NULL, 10);

    *dev = makedev(major, minor);
    return 1;
}

void nashBlockInit(nashContext *c)
{
    if (c->cache)
        return;

    if (blkid_get_cache(&c->cache, "/etc/blkid/blkid.tab") < 0)
        blkid_get_cache(&c->cache, NULL);
}

void nashBlockFinish(nashContext *c)
{
    blkid_put_cache(c->cache);
    c->cache = NULL;
}

void nash_bdev_destroy(struct nash_block_dev *dev)
{
    if (!dev)
        return;

    dev->type = 0;
    if (dev->sysfs_path) {
        free(dev->sysfs_path);
        dev->sysfs_path = NULL;
    }
    if (dev->dev_path) {
        free(dev->dev_path);
        dev->dev_path = NULL;
    }
    dev->devno = 0;
}

void nashBdevFreePtr(struct nash_block_dev **devp)
{
    struct nash_block_dev *dev;

    if (!devp)
        return;
    dev = *devp;
    nash_bdev_destroy(dev);
    free(dev);
    *devp = NULL;
}

int
nashBdevCmp(struct nash_block_dev *a, struct nash_block_dev *b)
{
    if (a->type != b->type)
        return b->type - a->type;
    return b->devno - a->devno;
}

struct nash_block_dev *
nashBdevDup(struct nash_block_dev *dev)
{
    struct nash_block_dev *new = NULL;
    if (!(new = calloc(1, sizeof (*new)))) {
err:
        nashBdevFreePtr(&new);
        return NULL;
    }
    if (!(new->sysfs_path = strdup(dev->sysfs_path)))
        goto err;
    if (!(new->dev_path = strdup(dev->dev_path)))
        goto err;

    new->type = dev->type;
    new->devno = dev->devno;
    return new;
}

struct nash_block_dev_iter {
    nashContext *nc;

    enum {
        ITERATING,
        POLLING,
        DONE
    } state;

    struct nash_list_iter *devlist_iter;

    struct timespec timeout;

    struct nash_block_dev bdev;
};

/*
 * This returns how many events were successfully processed.
 * Typically that means 1 if we got a block device, 0 if we didn't,
 * and -1 if a fatal error occurs.
 */
static int
block_try_uevent(nashContext *nc, nashUEvent *uevent,
        struct nash_block_dev **dev)
{
    char *slash = NULL;
    int maj, min;
    int rc = -1;
    struct nash_block_dev *bdev;
    char *path;
    char *tmp;

    tmp = envz_get(uevent->envz, uevent->envz_len, "SUBSYSTEM");
    if (!tmp || strcmp("block", tmp))
        return -1;

    if (!(tmp = envz_get(uevent->envz, uevent->envz_len, "MAJOR")))
        return -1;
    maj = atoi(tmp);

    if (!(tmp = envz_get(uevent->envz, uevent->envz_len, "MINOR")))
        return -1;
    min = atoi(tmp);

    if (!(tmp = envz_get(uevent->envz, uevent->envz_len, "DEVPATH")))
        return -1;

    if (!uevent->msg)
        return -1;

    if (!(bdev = calloc(1, sizeof (struct nash_block_dev))))
        return -1;

    if (!strcmp(uevent->msg, "add")) {
        bdev->type = ADD;
    } else if (!strcmp(uevent->msg, "remove")) {
        bdev->type = REMOVE;
        /* XXX we're not handling removes yet */
        goto err;
    } else {
        goto err;
    }

    bdev->devno = makedev(maj, min);

    if (asprintf(&bdev->sysfs_path, "/sys%s", tmp) < 0)
        goto err;

    slash = strrchr(tmp, '/');
    for (ssize_t n = MAX(MIN(strlen(slash), PATH_MAX), 0); n >= 0; n--)
        slash[n] = slash[n] == '!' ? '/' : slash[n];
    if (asprintf(&bdev->dev_path, "/dev%s", slash) < 0)
        goto err;

    smartmknod(bdev->dev_path, S_IFBLK | 0700, bdev->devno);
    path = nashDmDevGetName(nc, bdev->devno);
    if (path) {
        free(bdev->dev_path);
        bdev->dev_path = path;
        smartmknod(bdev->dev_path, S_IFBLK | 0700, bdev->devno);
    }

    *dev = bdev;
    return 0;
err:
    *dev = NULL;
    nashBdevFreePtr(&bdev);
    return rc;
}

static int block_process_one_uevent(nashContext *nc,
        struct timespec *timeout, struct nash_dev_node **nodep)
{
    struct nash_dev_node *node = NULL;
    nashUEvent uevent;
    struct nash_block_dev *dev = NULL;
    int rc;

    if (nodep)
        *nodep = NULL;

    memset(&uevent, '\0', sizeof (uevent));

    if ((rc = nashGetUEvent(nc, timeout, &uevent)) <= 0)
        return rc;

    if (block_try_uevent(nc, &uevent, &dev) >= 0) {
        node = nash_dev_tree_process_bdev(nc, dev);
        if (nodep)
            *nodep = node;
    }
    if (uevent.msg)
        free(uevent.msg);
    if (uevent.path)
        free(uevent.path);
    while (uevent.envz)
        argz_delete(&uevent.envz, &uevent.envz_len, uevent.envz);
    return rc;
}

static void block_process_pending_uevents(nashContext *nc)
{
    struct timespec timeout;

    do {
        usectospec(1, &timeout);
    } while (block_process_one_uevent(nc, &timeout, NULL) > 0);
}

nashBdevIter
nashBdevIterNewPoll(nashContext *nc, struct timespec *timeout)
{
    nashBdevIter iter = NULL;

    if (!nc->uh && !(nc->uh = nashUEventHandlerNew(nc)))
        return NULL;

    if (!nc->devices) {
        if (!(nc->devices = nash_dev_tree_alloc(nc)) < 0)
            return NULL;
    
        nash_dev_tree_populate_from_sysfs(nc);
        block_process_pending_uevents(nc);
    }

    if (!(iter = calloc(1, sizeof(*iter))))
        return NULL;

    iter->nc = nc;
    iter->state = ITERATING;

    if (timeout && !speczero(timeout)) {
        iter->timeout = *timeout;
    } else {
        usectospec(0, &iter->timeout);
    }

    return iter;
}

void
nashBdevIterEnd(nashBdevIter *iterp)
{
    nashBdevIter iter;
    if (!iterp || !*iterp)
        return;
    iter = *iterp;

    if (iter->devlist_iter)
        nash_list_iter_end(iter->devlist_iter);

    free(iter);
    *iterp = NULL;
}

int
nashBdevIterNext(nashBdevIter *iterp, struct nash_block_dev **dev)
{
    nashBdevIter iter;
    struct timespec timeout;
    int rc;

    if (!iterp || !*iterp)
        return 0;
    iter = *iterp;

    nash_bdev_destroy(&iter->bdev);

    timeout = iter->timeout;
    while (1) {
        struct nash_dev_node *node = NULL;

        switch (iter->state) {
            case ITERATING:
                iter->devlist_iter = nash_list_iter_next(
                    iter->nc->devices->devs, iter->devlist_iter);

                if (iter->devlist_iter) {
                    node = nash_list_iter_data(iter->devlist_iter);
                    *dev = node->bdev;
                    
                    return 1;
                }

                iter->state = POLLING;
                continue;
            case POLLING:
                timeout = iter->timeout;
                rc = block_process_one_uevent(iter->nc, &timeout, &node);
                if (rc > 0 && node) {
                    *dev = node->bdev;
                    return 1;
                }
                if (rc <= 0)
                    iter->state = DONE;
                continue;
            case DONE:
                nashBdevIterEnd(iterp);
                iter = NULL;
                return 0;
        }
    }
    /* we never get here */
    return -1;
}

char *
nashFindDeviceByDevno(nashContext *nc, dev_t devno)
{
    nashBdevIter biter;
    struct nash_block_dev *dev = NULL;
    char *path = NULL;

    biter = nashBdevIterNew(nc);
    while (nashBdevIterNext(&biter, &dev) > 0) {
        if (dev->devno == devno) {
            path = strdup(strrchr(dev->dev_path, '/')+1);
            break;
        }
    }
    nashBdevIterEnd(&biter);
    return path;
}

static char *
block_find_fs_by_keyvalue(nashContext *nc, const char *key, const char *value)
{
    nashBdevIter biter;
    struct nash_block_dev *dev = NULL;
    blkid_dev bdev = NULL;

    biter = nashBdevIterNew(nc);
    while (nashBdevIterNext(&biter, &dev) > 0) {
        blkid_tag_iterate titer;
        const char *type, *data;
        char *dmname = NULL, *name = NULL;

        if (!strncmp(dev->dev_path, "/dev/dm-", 8))
            dmname = nashDmDevGetName(nc, dev->devno);
        name = dmname ? dmname : dev->dev_path;
        bdev = blkid_get_dev(nc->cache, name, BLKID_DEV_NORMAL);
        if (dmname)
            free(dmname);
        if (!bdev)
            continue;
        titer = blkid_tag_iterate_begin(bdev);
        while (blkid_tag_next(titer, &type, &data) >= 0) {
            if (!strcmp(type, key) && !strcmp(data, value)) {
                name = strdup(blkid_dev_devname(bdev));
                blkid_tag_iterate_end(titer);
                nashBdevIterEnd(&biter);
                return name;
            }
        }
        blkid_tag_iterate_end(titer);
    }
    nashBdevIterEnd(&biter);

    return NULL;
}

char *
nashFindFsByLabel(nashContext *nc, const char *label)
{
    return block_find_fs_by_keyvalue(nc, "LABEL", label);
}

char *
nashFindFsByUUID(nashContext *nc, const char *uuid)
{
    return block_find_fs_by_keyvalue(nc, "UUID", uuid);
}

char *
nashFindFsByName(nashContext *nc, const char *name)
{
    blkid_dev bdev = NULL;
    nassert(name && name[0]);

    if (!access("/sys/block", F_OK)) {
        /* populate the whole cache */
        block_find_fs_by_keyvalue(nc, "unlikely","unlikely");

        /* now look our device up */
        bdev = blkid_get_dev(nc->cache, name, BLKID_DEV_NORMAL);
    }

    /* do not return immediately, check device access
       since blkid now returns cached devices that may not exist */
    if (bdev)
        name = blkid_dev_devname(bdev);

    if (!access(name, F_OK))
        return strdup(name);
    return NULL;
}

char *
nashAGetPathBySpec(nashContext *nc, const char * spec)
{
    char *path;

    nashBlockInit(nc);
    if (!strncmp(spec, "LABEL=", 6))
        path = nashFindFsByLabel(nc, spec+6);
    else if (!strncmp(spec, "UUID=", 5))
        path = nashFindFsByUUID(nc, spec+5);
    else
        path = nashFindFsByName(nc, spec);
    nashBlockFinish(nc);

    return path;
}

int
nashMkPathBySpec(nashContext *nc, const char *spec, const char *path)
{
    char *existing = nashGetPathBySpec(nc, spec);
    struct stat sb;

    if (!existing) {
        char *p;
        dev_t devNum = strtoul(spec, &p, 16);
        if (!*p && devNum > 0)
            return smartmknod(path, S_IFBLK | 0700, devNum);
    }

    if (!existing || stat(existing, &sb) < 0 || !S_ISBLK(sb.st_mode))
        return -1;

    return smartmknod(path, S_IFBLK | 0700, sb.st_rdev);
}

static int
block_disable_partition(int fd, int partno)
{
    struct blkpg_partition part = {
        .pno = partno,
    };
    struct blkpg_ioctl_arg io = {
        .op = BLKPG_DEL_PARTITION,
        .datalen = sizeof(part),
        .data = &part,
    };
    int ret;

    ret = ioctl(fd, BLKPG, &io);
    if (ret < 0)
        return 0;
    return 1;
}

int
nashDisablePartitions(const char *devname)
{
    int fd;
    int partno;
    char path[256];
    int ret = 0;

    snprintf(path, 255, "/dev/%s", devname);
    fd = open(path, O_RDWR);
    if (fd < 0)
        return fd;

    for (partno = 1; partno <= 256; partno++)
        ret += block_disable_partition(fd, partno);
    return ret ? ret : -1;
}

/*
 * vim:ts=8:sw=4:sts=4:et
 */
